import json

import betamax  # type: ignore
import pytest


config = betamax.Betamax.configure()
config.cassette_library_dir = "tests/fixtures/cassettes"
# only replay recorded cassettes -
# error if an actual HTTP request would be necessary
# this is to prevent packaging issues in the offline environment (like rpm build)
# change to 'once' to enable recording new cassettes when writing new tests
config.default_cassette_options["record_mode"] = "none"


@pytest.fixture(scope="session")
def fake_fedora_licenses():
    with open("tests/fedora_license_data.json", "r", encoding="utf-8") as f:
        return json.load(f)



# This script is a cut and modified version of pyreq2rpm.
#
# Copyright 2019 Gordon Messmer <gordon.messmer@gmail.com>
#
# Original: https://github.com/gordonmessmer/pyreq2rpm
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from packaging.version import parse

class RpmVersion():
    def __init__(self, version_id):
        version = parse(version_id)
        if isinstance(version._version, str):
            self.version = version._version
        else:
            self.epoch = version._version.epoch
            self.version = list(version._version.release)
            self.pre = version._version.pre
            self.dev = version._version.dev
            self.post = version._version.post
            # version.local is ignored as it is not expected to appear
            # in public releases
            # https://www.python.org/dev/peps/pep-0440/#local-version-identifiers

    def is_legacy(self):
        return isinstance(self.version, str)

    def __str__(self):
        if self.is_legacy():
            return self.version
        if self.epoch:
            rpm_epoch = str(self.epoch) + ':'
        else:
            rpm_epoch = ''
        rpm_version = '.'.join(str(x) for x in self.version)
        if self.pre:
            rpm_suffix = '~{}'.format(''.join(str(x) for x in self.pre))
        elif self.dev:
            rpm_suffix = '~~{}'.format(''.join(str(x) for x in self.dev))
        elif self.post:
            rpm_suffix = '^post{}'.format(self.post[1])
        else:
            rpm_suffix = ''
        return '{}{}{}'.format(rpm_epoch, rpm_version, rpm_suffix)



import pytest

from pyp2spec.utils import filter_license_classifiers, prepend_name_with_python
from pyp2spec.utils import normalize_name, get_extras, is_archful
from pyp2spec.utils import normalize_as_wheel_name, archive_name
from pyp2spec.utils import resolve_url, SdistNotFoundError, MissingPackageNameError
from pyp2spec.utils import create_compat_name


def test_license_classifier_read_correctly():
    fake_pkg_data = [
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Operating System :: POSIX :: Linux",
        "License :: OSI Approved :: MIT License",
        "License :: OSI Approved :: MIT No Attribution License (MIT-0)",
        "Development Status :: 3 - Alpha",
    ]

    assert filter_license_classifiers(fake_pkg_data) == [
        "License :: OSI Approved :: MIT License",
        "License :: OSI Approved :: MIT No Attribution License (MIT-0)"
    ]


@pytest.mark.parametrize(
    ("pypi_name", "alt_version", "expected"), [
        ("foo", "3.10", "python3.10-foo"),
        ("python-foo", "3.9", "python3.9-foo"),
        ("python_foo", "3.12", "python3.12-foo"),
        ("foo", None, "python-foo"),
        ("python-foo", None, "python-foo"),
        ("python_foo", None, "python-foo"),
    ]
)
def test_python_name(pypi_name, alt_version, expected):
    assert prepend_name_with_python(normalize_name(pypi_name), alt_version) == expected


def test_extras_detected_correctly_from_requires_dist():
    requires_dist = [
        "foo ; platform_system == 'Windows'",
        "bar ; python_version < '3.8'",
        "baz",
        "foobar ; extra == 'docs'",
        "foobaz>=5.3.1",
        "spam>=3.5.0 ; extra == 'lint'",
        "ham[eggs]>=0.10; extra == 'test'",
        "eggs>=2.12; implementation_name != 'pypy' and extra == 'dev'",
    ]

    assert get_extras([], requires_dist) == ["dev", "docs", "lint", "test"]


def test_extras_detected_correctly_from_provides_extra():
    provides_extra = ['docs', 'lint', 'dev', 'foo']
    requires_dist = [
        "foo ; platform_system == 'Windows'",
        "bar ; python_version < '3.8'",
        "baz",
        "foobar ; extra == 'docs'",
        "foobaz>=5.3.1",
        "spam>=3.5.0 ; extra == 'lint'",
        "ham[eggs]>=0.10; extra == 'test'",
        "eggs>=2.12; implementation_name != 'pypy' and extra == 'dev'",
    ]
    assert get_extras(provides_extra, requires_dist) == ["dev", "docs", "foo", "lint"]


def test_extras_detected_correctly_from_provides_extra_no_requires():
    provides_extra = ['docs', 'lint', 'dev', 'foo']
    # this is improbable, but let's be sure we read from the provides_extra without issues
    requires_dist = []
    assert get_extras(provides_extra, requires_dist) == ["dev", "docs", "foo", "lint"]


@pytest.mark.parametrize(
    ("wheel_name", "archful"), [
        ("sampleproject-3.0.0-py3-none-any.whl", False),
        ("numpy-1.26.0-cp39-cp39-win_amd64.whl", True),
        ("numpy-1.26.0-cp312-cp312-manylinux_2_17_x86_64.manylinux2014_x86_64.whl", True),
        ("numpy-1.26.0-cp39-cp39-musllinux_1_1_x86_64.whl", True),
    ]
)
def test_archfulness_is_detected(wheel_name, archful):
    urls = [
        {
            "packagetype": "bdist_wheel",
            "filename": wheel_name,
        },
    ]
    assert is_archful(urls) is archful


def test_archfulness_is_detected_from_multiple_urls_1():
    urls = [
        {
            "packagetype": "sdist",
            "filename": "sampleproject-3.0.0.tar.gz",
        },
        {
            "packagetype": "bdist_wheel",
            "filename": "sampleproject-3.0.0-py3-none-any.whl",
        },
    ]
    assert not is_archful(urls)


def test_archfulness_is_detected_from_multiple_urls_2():
    urls = [
        {
            "packagetype": "bdist_wheel",
            "filename": "sampleproject-3.0.0-py3-none-any.whl",
        },
        {
            "packagetype": "bdist_wheel",
            "filename": "sampleproject-3.0.0-cp312-abi3-manylinux1_x86_64.whl",
        },
    ]
    assert is_archful(urls)



def test_archfulness_is_detected_from_multiple_urls_3():
    urls = [
        {
            "packagetype": "bdist_wheel",
            "filename": "sampleproject-3.0.0-cp312-abi3-manylinux1_x86_64.whl",
        },
        {
            "packagetype": "bdist_wheel",
            "filename": "sampleproject-3.0.0-cp39-cp39-win_amd64.whl",
        },
        {
            "packagetype": "bdist_wheel",
            "filename": "sampleproject-3.0.0-py3-none-any.whl",
        },
    ]
    assert is_archful(urls)


@pytest.mark.parametrize(
    ("name", "expected"), [
        ("my-package-foo", "my_package_foo"),
        ("my.package", "my_package"),
        ("my_package-", "my_package_"),
        ("noweirdchars", "noweirdchars"),
    ]
)
def test_normalize_as_wheel_name(name, expected):
    assert normalize_as_wheel_name(name) == expected


def test_empty_package_name_results_in_exception():
    with pytest.raises(MissingPackageNameError):
        normalize_as_wheel_name("")


def test_archive_name_valid():
    archive_urls = [
        {
            "packagetype": "bdist_wheel",
            "filename": "example-1.0.0-py3-none-any.whl"
        },
        {
            "packagetype": "sdist",
            "filename": "example-1.0.0.tar.gz"
        },
    ]
    assert archive_name(archive_urls) == "example-1.0.0.tar.gz"


def test_archive_name_no_sdist():
    archive_urls = [
        {
            "packagetype": "bdist_wheel",
            "filename": "example-1.0.0-py3-none-any.whl"
        },
    ]
    with pytest.raises(SdistNotFoundError):
        archive_name(archive_urls)


def test_archive_name_empty_list():
    with pytest.raises(SdistNotFoundError):
        archive_name([])


def test_project_urls_valid_single():
    urls = {"homepage": "https://example.com"}
    assert resolve_url(urls) == "https://example.com"


def test_project_urls_valid_multiple():
    urls = {
        "homepage": "https://example.com",
        "documentation": "https://docs.example.com",
    }
    assert resolve_url(urls) == "https://example.com"


def test_project_urls_empty():
    assert resolve_url({}) == "..."


@pytest.mark.parametrize(
    ("name", "compat", "expected"), [
        ("my-package-foo", None, "my-package-foo"),
        ("my-package-foo", "3.5", "my-package-foo3.5"),
        ("my-package-foo2", None, "my-package-foo2"),
        ("my-package-foo2", "3.5", "my-package-foo2_3.5"),
    ]
)
def test_create_compat_name(name, compat, expected):
    assert create_compat_name(name, compat) == expected



"""Test the functionality of conf2spec part of the tool.
conf2spec can do a bit more that the fully automatic generation of spec files.
The tests test both the automatically generated files and the custom
configurations.
"""

from pathlib import Path

import pytest

from pyp2spec import conf2spec


@pytest.fixture
def config_dir():
    return "tests/test_configs/"


@pytest.mark.parametrize(
    ("conf", "expected"), [
        ("default_python-click.conf", False),
        ("customized_markdown-it-py.conf", True),
    ]
)
def test_automode_flag_is_loaded(config_dir, conf, expected):
    config_path = config_dir + conf
    config = conf2spec.ConfigFile(conf2spec.load_config_file(config_path))
    assert config.get_bool("automode") == expected



@pytest.mark.parametrize(
    ("conf", "expected"), [
        ("default_python-click.conf", False),
        ("default_python-numpy.conf", True),
    ]
)
def test_archful_flag_is_loaded(config_dir, conf, expected):
    config_path = config_dir + conf
    config = conf2spec.ConfigFile(conf2spec.load_config_file(config_path))
    assert config.get_bool("archful") is expected


@pytest.mark.parametrize(
    ("conf", "db"), [
        ("default_python-click.conf", False),  # default, noarch, no quirks
        ("customized_markdown-it-py.conf", False),  # automode on
        ("customized_python-sphinx.conf", False),  # contains extras
        ("default_python-numpy.conf", False),  # archful
        ("default_python3.9-pello.conf", False),  # custom Python version
        ("default_python-pytest7.2.conf", False),  # compat version
        ("default_python-pytest7.conf", False),  # compat version - lower granularity
        ("default_python-urllib3_2.conf", False), # compat version - pkgname with a digit
        ("default_python-pello.conf", True),  # declarative build system
    ]
)
def test_default_generated_specfile(file_regression, config_dir, conf, db):
    # Run the conf2spec converter
    rendered_file = conf2spec.create_spec_file(config_dir + conf, {"declarative_buildsystem": db})

    # Compare the results
    with open(rendered_file, "r", encoding="utf-8") as rendered_f:
        rendered = rendered_f.read()
    try:
        file_regression.check(
            rendered,
            fullpath=f"tests/expected_specfiles/{rendered_file}",
        )
    finally:
        # Cleanup - remove created file
        Path.unlink(Path(rendered_file))


@pytest.mark.parametrize(
    ("pypi_version", "rpm_version"), [
        ("1.1.0", "1.1.0"),
        ("1!0.2.13", "1:0.2.13"),
        ("0.0.2-beta1", "0.0.2~b1"),
        ("0.5.40-0", "0.5.40^post0"),
    ]
)
def test_pypi_version_is_converted_to_rpm(pypi_version, rpm_version):
    assert conf2spec.convert_version_to_rpm_scheme(pypi_version) == rpm_version


@pytest.mark.parametrize(
    ("version", "expected"), [
        ("1.2-3", False),
        ("1.2.3", True),
    ]
)
def test_source_macro(version, expected):
    assert conf2spec.same_as_rpm(version) is expected


def test_additional_build_requires_if_archful():
    fake_config_data = {"archful": True}
    fake_config = conf2spec.ConfigFile(fake_config_data)
    assert "gcc" in conf2spec.list_additional_build_requires(fake_config)


def test_additional_build_requires_if_noarch():
    fake_config_data = {"archful": False}
    fake_config = conf2spec.ConfigFile(fake_config_data)
    assert "gcc" not in conf2spec.list_additional_build_requires(fake_config)


@pytest.mark.parametrize(
    ("version", "expected"), [
        ("4.2", "%{python3_pkgversion}"),
        ("3.15", "%{python3_pkgversion}"),
        ("", "3"),
    ]
)
def test_python3_pkgversion_or_3(version, expected):
    fake_config_data = {"python_alt_version": version}
    fake_config = conf2spec.ConfigFile(fake_config_data)
    assert conf2spec.python3_pkgversion_or_3(fake_config) == expected


def test_empty_license_string():
    fake_config_data = {"archful": False}
    fake_config = conf2spec.ConfigFile(fake_config_data)
    license_data = conf2spec.get_license_string(fake_config)
    assert license_data[0] == "..."
    assert license_data[1] == "# No license information obtained, it's up to the packager to fill it in"


def test_license_strings():
    fake_config_data = {"license": "MIT"}
    fake_config = conf2spec.ConfigFile(fake_config_data)
    license_data = conf2spec.get_license_string(fake_config)
    assert license_data[0] == "MIT"
    assert "# Check if the automatically generated License and its spelling is correct for Fedora" in license_data[1] 


@pytest.mark.parametrize(
    ("version", "extension", "expected"), [
        ("1.2", "tar.gz", "%{pypi_source foo}"),
        ("1.2", "zip", "%{pypi_source foo %{version} zip}"),
        ("1.2-3", "tar.gz", "%{pypi_source foo 1.2-3}"),
        ("0.0.2-beta1", "zip", "%{pypi_source foo 0.0.2-beta1 zip}"),
    ]
)
def test_source(version, extension, expected):
    fake_config_data = {"pypi_name": "foo", "archive_name": f"foo-{version}.{extension}", "source": "PyPI"}
    fake_config = conf2spec.ConfigFile(fake_config_data)
    assert conf2spec.source(fake_config, version) == expected


@pytest.mark.parametrize(
    ("version", "expected"), [
        ("1.2", "%{version}"),
        ("1.2-3", "1.2-3"),
        ("0.0.2-beta1", "0.0.2-beta1"),
    ]
)
def test_pypi_version_or_macro(version, expected):
    assert conf2spec.pypi_version_or_macro(version) == expected



from __future__ import annotations
from dataclasses import dataclass, asdict, field
from functools import wraps
import sys

import click
import tomli_w
from packaging.metadata import RawMetadata
from requests import Session

from pyp2spec.license_processor import check_compliance, resolve_license_expression
from pyp2spec.utils import Pyp2specError, normalize_name, get_extras, get_summary_or_placeholder
from pyp2spec.utils import prepend_name_with_python, archive_name
from pyp2spec.utils import is_archful, resolve_url, create_compat_name
from pyp2spec.utils import warn, caution, inform, yay
from pyp2spec.pypi_loaders import load_from_pypi, load_core_metadata_from_pypi, CoreMetadataNotFoundError


@dataclass
class PackageInfo:
    pypi_name: str
    pypi_version: str
    summary: str
    url: str
    extras: list
    license_files_present: bool
    license: str | None = None
    source: str = "PyPI"  # only one default source now
    # These are added later, when the object is already constructed:
    python_name: str = field(init=False)
    archful: bool = field(init=False)
    archive_name: str = field(init=False)
    # These may or may not be ever set
    python_alt_version: str | None = field(default=None)
    automode: bool | None = field(default=None)
    compat: str | None = field(default=None)


def is_package_name(package: str) -> bool:
    """Least-effort check whether `package` is a package name or URL.
    Canonical package names can't contain '/'.
    """

    return "/" not in package


def prepare_package_info(data: RawMetadata | dict) -> PackageInfo:
    if not (project_urls := data.get("project_urls", {})):
        if (homepage := data.get("home_page", "")):
            project_urls["home_page"] = homepage
    return PackageInfo(
        pypi_name=normalize_name(data.get("name", "")),
        pypi_version=data.get("version", ""),
        summary=get_summary_or_placeholder(data.get("summary", "")),
        url=resolve_url(project_urls),
        extras=get_extras(data.get("provides_extra", []), data.get("requires_dist", [])),
        license_files_present=bool(data.get("license_files")),
        license=resolve_license_expression(data)
    )


def add_archive_data(pkg: PackageInfo, pypi_data: dict) -> PackageInfo:
    """We can obtain the archive information from PyPI only."""
    pkg.archful = is_archful(pypi_data["urls"])
    pkg.archive_name = archive_name(pypi_data["urls"])
    return pkg


def gather_package_info(core_metadata: RawMetadata | None, pypi_package_data: dict) -> PackageInfo:
    if core_metadata is not None:
        pkg = prepare_package_info(core_metadata)
    else:
        pkg = prepare_package_info(pypi_package_data["info"])
    pkg = add_archive_data(pkg, pypi_package_data)
    return pkg


def create_package_from_source(
    package: str,
    version: str | None,
    compat: str | None,
    session: Session | None
) -> PackageInfo:
    """Determine the best source for the given package name and create a PackageInfo instance.
    """
    if is_package_name(package):
        # explicit `session` argument is needed for testing
        pypi_pkg_data = load_from_pypi(package, version=version,
        compat=compat, session=session)
        try:
            core_metadata = load_core_metadata_from_pypi(pypi_pkg_data, session=session)
        # if no core metadata found, we will fall back to PyPI API
        except CoreMetadataNotFoundError:
            core_metadata = None
    else:
        raise NotImplementedError("pyp2spec can't currently handle URLs.")
    # The processed package info is the basis for config contents
    return gather_package_info(core_metadata, pypi_pkg_data)


def create_config_contents(
    options: dict[str, Any],
    session: Session | None = None
) -> dict:
    """Use `package` and provided options to create the whole config contents.
    Return pkg_info dictionary.
    """

    package = options.get("package")
    version = options.get("version")
    compat = options.get("compat")
    pkg_info = create_package_from_source(package, version, compat, session)

    python_alt_version = options.get("python_alt_version")
    pkg_info.python_name = prepend_name_with_python(pkg_info.pypi_name, python_alt_version)

    if compat is not None:
        inform(f"Creating a compat package for version: '{compat}'")
        pkg_info.compat = compat

    if version is None:
        inform(f"Assuming the version found on PyPI: '{pkg_info.pypi_version}'")

    if pkg_info.archful:
        caution("Package contains compiled extensions - you may need to specify additional build requirements")

    if pkg_info.license is None:
        warn("WARNING: No valid license found. Inspect the project manually to find the license")
    else:
        if options.get("fedora_compliant"):
            is_compliant, results = check_compliance(pkg_info.license, session=session)
            if not is_compliant:
                warn(f"The license '{pkg_info.license}' is not compliant with Fedora")
            if results["bad"]:
                info_str = "Found identifiers: '{0}' aren't allowed in Fedora."
                warn(info_str.format(", ".join(results["bad"])))
            if results["good"]:
                info_str = "Found identifiers: '{0}' are good for Fedora."
                inform(info_str.format(", ".join(results["good"])))

    # Configuration settings
    if python_alt_version is not None:
        inform(f"Assuming build for Python: {python_alt_version}")
        pkg_info.python_alt_version = python_alt_version

    if options.get("automode"):
        pkg_info.automode = True

    pkg_dict = asdict(pkg_info)
    # sort the dictionary alphabetically for output consistency
    # only keep the values that aren't None - TOML can't handle that
    return {key: pkg_dict[key] for key in sorted(pkg_dict.keys()) if pkg_dict[key] is not None}


def save_config(contents: dict, output: str | None = None) -> str:
    """Write config file to a given destination.
    If none is provided, save it to current directory with package name as file name.
    Return the saved file name.
    """
    if not output:
        package_name = create_compat_name(contents.get("python_name"), contents.get("compat"))
        output = f"{package_name}.conf"
    with open(output, "wb") as f:
        tomli_w.dump(contents, f, multiline_strings=True)
    yay(f"Configuration file was saved successfully to '{output}'")
    return output


def create_config(options: dict) -> str:
    """Create and save config file."""

    contents = create_config_contents(options)
    return save_config(contents, options["config_output"])


def pypconf_args(func):  # noqa
    @click.argument("package")
    @click.option(
        "--config-output", "-c",
        help="Provide custom output for configuration file",
    )
    @click.option(
        "--version", "-v",
        help="Provide package version to query PyPI for, default: latest",
    )
    @click.option(
        "--fedora-compliant", is_flag=True,
        help="Check whether license is compliant with Fedora",
    )
    @click.option(
        "--automode", "-a", is_flag=True,
        help="Enable buildability of the generated spec in automated environments",
    )
    @click.option(
        "--python-alt-version", "-p",
        help="Provide specific Python version to build for, e.g 3.11",
    )
    @click.option(
        "--compat",
        help="Create a compat package for a given version",
    )
    @wraps(func)
    def wrapper(*args, **kwargs): # noqa
        return func(*args, **kwargs)
    return wrapper

@click.command()
@pypconf_args
def main(**options):  # noqa
    try:
        create_config(options)
    except (Pyp2specError, NotImplementedError) as exc:
        warn(f"Fatal exception occurred: {exc}")
        sys.exit(1)

if __name__ == "__main__":
    main()



"""
This module contains common functions and classes imported in multiple other modules.
"""
from __future__ import annotations

import re
import string
from functools import partial

import click

from packaging.requirements import Requirement


warn = partial(click.secho, fg="red")
caution = partial(click.secho, fg="magenta")
inform = partial(click.secho, fg="yellow")
yay = partial(click.secho, fg="green")


class Pyp2specError(Exception):
    """Metaexception to derive the custom errors from"""


class SdistNotFoundError(Pyp2specError):
    """Raised when there's no sdist file in the PyPI metadata"""


class MissingPackageNameError(Pyp2specError):
    """Raised when there's no package name in the metadata"""


def normalize_name(package_name: str) -> str:
    """Normalize given package name as defined in PEP 503.
    The resulting string better conforms with Fedora's Packaging Guidelines."""

    if not package_name:
        raise MissingPackageNameError("Cannot create a package without a name")
    return re.sub(r"[-_.]+", "-", package_name).lower()


def normalize_as_wheel_name(package_name: str) -> str:
    """Normalize as in the wheel specification:
    https://packaging.python.org/en/latest/specifications/binary-distribution-format/#escaping-and-unicode
    PEP 625 specifies sdist names to this format."""

    return normalize_name(package_name).replace("-", "_")


def prepend_name_with_python(name: str, python_alt_version: str | None = None) -> str:
    """Create a component name for the specfile.

    Prepend the name with 'python' (unless it already starts with it).
    Add the Python alternative version, if it's defined.
    Valid outcomes:
    package name: foo, python_alt_version: 3.11
    -> python3.11-foo

    package name: foo, python_alt_version: None
    -> python-foo

    package name: python-foo, python_alt_version: 3.12
    -> python3.12-foo

    package name: python-foo, python_alt_version: None
    -> python-foo
    """

    alt_version = "" if python_alt_version is None else python_alt_version
    if name.startswith("python"):
        return name.replace("python", f"python{alt_version}")
    return f"python{alt_version}-{name}"


def filter_license_classifiers(classifiers_list: list) -> list:
    """Return the list of license classifiers defined for the package.

    Filter out the parent categories `OSI-/DFSG Approved` which don't have any meaning.
    """

    return  [
        c for c in classifiers_list
        if (
            c.startswith("License")
            and c not in ("License :: OSI Approved", "License :: DFSG approved")
        )
    ]


def get_summary_or_placeholder(summary: str) -> str:
    """Return either a summary or a "..." string.

    Summary is an optional field, so it may be empty or it can consist of
    multi-line strings which we can't use.
    """

    if not summary or len(summary.split("\n")) > 1:
        summary = "..."
    return summary


def get_extras(provides_extra: list, requires_dist: list) -> list:
    """Return the sorted list of the found extras names.

    Packages define extras explicitly via `Provides-Extra` and
    indirectly via `Requires-Dist` metadata.
    Try to read them from `provides_extra` first.
    PyPI metadata doesn't publish `provides_extra` for the older releases,
    but it is possible to derive extras names from the `requires_dist` key.
    Example value of `requires_dist`:
    ["sphinxcontrib-websupport ; extra == 'docs'", "flake8>=3.5.0 ; extra == 'lint'"]
    If a package defines an extra with no requirements,
    we can't detect that from requires_dist.
    """
    if provides_extra:
        return sorted(provides_extra)

    extra_from_req = re.compile(r'''\bextra\s+==\s+["']([^"']+)["']''')
    extras = set()
    if requires_dist:
        for required_dist in requires_dist:
            # packaging.Requirement can parse the markers, but it
            # doesn't provide their string representations,
            # hence we need to use regex to pick them out
            req = Requirement(required_dist)
            if found := re.search(extra_from_req, str(req.marker)):
                extras.add(found.group(1))
    return sorted(extras)


def archive_name(archive_urls: list) -> str:
    """Return the given's package version sdist name for further processing.
    Quit the script if not found (bdists can't be processed).
    """
    for entry in archive_urls:
        if entry["packagetype"] == "sdist":
            return entry["filename"]
    raise SdistNotFoundError("Sdist not found, valid spec file cannot be produced")


def is_archful(archive_urls: list) -> bool:
    """Determine if package is archful by checking the wheel filenames.

    Wheel name consists of defined fields, one of them being an abi tag.
    Example abi tags:
    - click-0.2-py2.py3-none-any.whl -> "none"
    - cryptography-2.2-cp34-abi3-manylinux1_x86_64.whl -> "abi3"
    If the value is "none", wheel was not built for a specific architecture,
    probably containing pure Python modules.
    Other values indicate build for an architecture, which can mean
    the presence of compiled extensions.
    Packages can publish multiple wheels, the pure-Python alongside the compiled ones.
    For our purposes, if we find at least one wheel with an abi tag different that "none",
    we consider the package archful.
    The compiled extensions bring optimizations and in Fedora,
    it is generally encouraged to bring in the optional features of the packages.
    """

    for entry in archive_urls:
        if entry["packagetype"] == "bdist_wheel":
            abi_tag = entry["filename"].split("-")[-2]
            if abi_tag != "none":
                return True
    # all of the found wheel names had 'none' as abi_tag
    return False


def _normalize_url_label(label: str) -> str:
    """
    Copied verbatim from
    https://packaging.python.org/en/latest/specifications/well-known-project-urls/#label-normalization
    """
    chars_to_remove = string.punctuation + string.whitespace
    removal_map = str.maketrans("", "", chars_to_remove)
    return label.translate(removal_map).lower()


def resolve_url(urls: dict) -> str:
    """
    Project urls are an optional field and come in a form of a dict, e.g.:
    {"homepage": "https://mypackagehomepage.com"}.
    There may be multiple urls defined and keys are arbitrary.
    Try to return url for normalized "homepage", then "source" keys.
    Otherwise return the first value from the dictionary or "..." if the dict is empty.
    """
    normalized_urls = {_normalize_url_label(k): v for k, v in urls.items()}
    well_known_urls = normalized_urls.get("homepage") \
        or normalized_urls.get("source")\
        or normalized_urls.get("sourcecode")\
        or normalized_urls.get("github")\
        or normalized_urls.get("repository")
    if well_known_urls:
        return well_known_urls
    else:
        if values_list := list(normalized_urls.values()):
            return values_list[0]
        return "..."


def create_compat_name(name: str, compat: str | None) -> str:
    if not compat:
        return name
    if name[-1].isdigit():
        return f"{name}_{compat}"
    return f"{name}{compat}"



import sys

import click

from pyp2spec.pyp2conf import create_config, pypconf_args
from pyp2spec.conf2spec import create_spec_file
from pyp2spec.utils import Pyp2specError
from pyp2spec.utils import warn


@click.command()
@pypconf_args
@click.option(
    "--spec-output", "-o",
    help="Provide custom output where spec file will be saved",
)
@click.option(
    "--declarative-buildsystem", is_flag=True, default=False,
    help="Create a spec file with pyproject declarative buildsystem (experimental)",
)
def main(**options):  # noqa
    try:
        if options["automode"] and options["declarative_buildsystem"]:
            raise Pyp2specError("Declarative buildsystem doesn't work with automode")
        config_file = create_config(options)
        create_spec_file(config_file, options)
    except (Pyp2specError, NotImplementedError) as exc:
        warn(f"Fatal exception occurred: {exc}")
        sys.exit(1)


if __name__ == "__main__":
    main()



from __future__ import annotations

import sys

from importlib.resources import files
from typing import Any

import click

try:
    import tomllib
except ImportError:
    import tomli as tomllib  # type: ignore

from jinja2 import Template

from pyp2spec.rpmversion import RpmVersion
from pyp2spec.utils import Pyp2specError, create_compat_name
from pyp2spec.utils import warn, yay


TEMPLATE_FILENAME = "template.spec"
ADDITIONAL_BUILD_REQUIRES_ARCHFUL = ["gcc"]
ADDITIONAL_BUILD_REQUIRES_NOARCH: list[str] = []


class ConfigError(Pyp2specError):
    """Raised when a config value has got a wrong type"""


class ConfigFile:
    """Validate configuration data."""

    def __init__(self, contents: dict) -> None:
        self.contents = contents

    def get_string(self, key: str) -> str:
        """Return a value for given key. Validate the value is a string.
        Raise ConfigError otherwise."""

        return self._get_value(key, str)

    def get_list(self, key: str) -> str:
        """Return a value for given key. Validate the value is a list.
        Raise ConfigError otherwise."""

        return self._get_value(key, list)

    def get_bool(self, key: str) -> str:
        """Return a value for given key. Validate the value is a boolean.
        Raise ConfigError otherwise."""

        return self._get_value(key, bool)

    def _get_value(self, key: str, val_type: type) -> Any:
        val = self.contents.get(key, val_type())
        if not isinstance(val, val_type):
            raise ConfigError(f"{val} must be a {val_type}")
        return val


def load_config_file(filename: str) -> dict:
    """Return loaded TOML configuration file contents."""

    with open(filename, "rb") as configuration_file:
        return tomllib.load(configuration_file)


def list_additional_build_requires(config: ConfigFile) -> list[str]:
    """Returns a list of additionally defined BuildRequires,

    The list differs for packages that are or aren't archful.
    """
    if config.get_bool("archful"):
        return ADDITIONAL_BUILD_REQUIRES_ARCHFUL
    return ADDITIONAL_BUILD_REQUIRES_NOARCH


def python3_pkgversion_or_3(config: ConfigFile) -> str:
    return "%{python3_pkgversion}" if config.get_string("python_alt_version") else "3"


def get_license_string(config: ConfigFile) -> tuple[str, str]:
    none_notice = "# No license information obtained, it's up to the packager to fill it in"
    detected_notice = ("# Check if the automatically generated License and its "
        "spelling is correct for Fedora\n"
        "# https://docs.fedoraproject.org/en-US/packaging-guidelines/LicensingGuidelines/")
    if (license := config.get_string("license")):
        return (license, detected_notice)
    return ("...", none_notice)


def convert_version_to_rpm_scheme(version: str) -> str:
    """If version follows PEP 440, return its value converted to RPM scheme.

    PEP 440: https://www.python.org/dev/peps/pep-0440/
    If the package uses a different versioning scheme (i.e. LegacyVersion),
    the returned value will be the same as the input one.
    Such value may or may not work with RPM.
    Automatic conversion of the LegacyVersions is not feasible, as stated in:
    https://lists.fedoraproject.org/archives/list/python-devel@lists.fedoraproject.org/message/5MGEHMTKOKR5U7ACIMUDRBKMSP6Y5NQD/
    """
    return str(RpmVersion(version))


def same_as_rpm(pypi_version: str) -> bool:
    return pypi_version == convert_version_to_rpm_scheme(pypi_version)


def archive_basename(config: ConfigFile, pypi_version: str) -> str:
    """Return the archive basename.

    The filename has been standardized in PEP 625, but many projects out there
    use the legacy, pre-standard, custom naming. Extract the basename the same
    way it's spelled in the archive name.

    Example: "My-package_Archive-1.0.4-12.tar.gz" -> "My-package_Archive"
    """
    filename = config.get_string("archive_name")

    # First, strip the suffix
    for suffix in (".tar.gz", ".zip"):
        filename = filename.removesuffix(suffix)
    # Second, get rid of the version string and the delimiter "-"
    return filename.replace("-" + pypi_version, "")


def is_zip(config: ConfigFile) -> bool:
    """
    The archive format standardized in PEP 625 is tar.gz,
    however some projects still use the legacy zip.
    If so, it has to be explicitly declared as an argument of %{pypi_source}.
    """
    return config.get_string("archive_name").endswith(".zip")


def source(config: ConfigFile, pypi_version: str) -> str:
    """Return valid pypi_source RPM macro.

    %pypi_source takes three optional arguments:
    <name>, <version> and <file_extension>.
    <name> is always passed: "%{pypi_source foo}".
    <version> is passed if PyPI's and RPM's version strings differ:
    "%{pypi_source foo 1.2-3}". If they're the same and the file extension
    is not zip, version is not passed.
    If archive is a zip file, %{pypi_source} must take all three args:
    "%{pypi_source foo %{version} zip}", "{pypi_source foo 1.2-3 zip}"
    """
    if config.get_string("source") == "PyPI":
        version_str = pypi_version_or_macro(pypi_version)
        basename = archive_basename(config, pypi_version)
        source_macro_args = basename

        if is_zip(config):
            source_macro_args += f" {version_str} zip"
        else:
            if version_str == pypi_version:
                source_macro_args += f" {version_str}"
        return "%{pypi_source " + source_macro_args + "}"
    else:
        raise NotImplementedError("pyp2spec can only construct %pypi_source macros")


def pypi_version_or_macro(pypi_version: str) -> str:
    if same_as_rpm(pypi_version):
        return "%{version}"
    return pypi_version


def fill_in_template(config: ConfigFile, declarative_buildsystem: bool) -> str:
    """Return template rendered with data from config file."""

    with (files("pyp2spec") / TEMPLATE_FILENAME).open("r", encoding="utf-8") as f:
        spec_template = Template(f.read())

    license, license_notice = get_license_string(config)

    pypi_version = config.get_string("pypi_version")

    result = spec_template.render(
        additional_build_requires=list_additional_build_requires(config),
        archful=config.get_bool("archful"),
        archive_name=archive_basename(config, pypi_version),
        automode=config.get_bool("automode"),
        compat=config.get_string("compat"),
        compat_name=create_compat_name(config.get_string("pypi_name"), config.get_string("compat")),
        declarative_buildsystem=declarative_buildsystem,
        extras=",".join(config.get_list("extras")),
        license=license,
        license_notice=license_notice,
        mandate_license=config.get_bool("license_files_present"),
        name=config.get_string("pypi_name"),
        python_compat_name=create_compat_name(config.get_string("python_name"), config.get_string("compat")),
        pypi_version=pypi_version_or_macro(pypi_version),
        python_alt_version=config.get_string("python_alt_version"),
        source=source(config, pypi_version),
        summary=config.get_string("summary"),
        test_top_level=config.get_bool("test_top_level"),
        python3_pkgversion=python3_pkgversion_or_3(config),
        url=config.get_string("url"),
        version=convert_version_to_rpm_scheme(pypi_version),
    )

    return result


def save_spec_file(config: ConfigFile, options: dict[str, Any]) -> str:
    """Save the spec file in the current directory if custom output is not set.
    Return the saved file name."""

    declarative_buildsystem = options.get("declarative_buildsystem", False)
    result = fill_in_template(config, declarative_buildsystem)
    output = options.get("spec_output")
    if output is None:
        output = create_compat_name(config.get_string("python_name"), config.get_string("compat"))
        output += ".spec"
    with open(output, "w", encoding="utf-8") as spec_file:
        spec_file.write(result)
    yay(f"Spec file was saved successfully to '{output}'")
    return output


def create_spec_file(config_file: str, options: dict[str, Any]) -> str | None:
    """Create and save the generate spec file."""
    config = ConfigFile(load_config_file(config_file))
    return save_spec_file(config, options)


@click.command()
@click.argument("config")
@click.option(
    "--spec-output",
    "-o",
    help="Provide custom output for spec file",
)
@click.option(
    "--declarative-buildsystem", is_flag=True, default=False,
    help="Create a spec file with pyproject declarative buildsystem (experimental)",
)
def main(config: str, **options: dict[str, Any]) -> None:
    try:
        create_spec_file(config, options)
    except (Pyp2specError, NotImplementedError) as exc:
        warn(f"Fatal exception occurred: {exc}")
        sys.exit(1)

if __name__ == "__main__":
    main()



from __future__ import annotations
from importlib.resources import files
from typing import Any
import json

from packaging.metadata import RawMetadata
from requests import Session
from license_expression import get_spdx_licensing, ExpressionError  # type: ignore

from pyp2spec.utils import Pyp2specError, filter_license_classifiers


TROVE2FEDORA_MAP: dict[str, str | None] = {}
FEDORA_LICENSES: dict[int, Any] = {}


class NoSuchClassifierError(Pyp2specError):
    """Raised when the detected license classifier doesn't exist in pyp2spec's data"""



def _load_package_resource(filename: str) -> dict[Any, Any]:
    with (files("pyp2spec") / filename).open("r", encoding="utf-8") as f:
        return json.load(f)


def _load_from_drive(source_path: str) -> dict[Any, Any]:
    with open(source_path, "r", encoding="utf-8") as f:
        return json.load(f)


def _load_from_url(url: str, session: Session | None=None) -> dict[Any, Any]:
    s = session or Session()
    response = s.get(url)
    response.raise_for_status()
    return response.json()


def classifiers_to_spdx_identifiers(classifiers: list) -> list | None:
    """Return the list of SPDX identifiers converted from the license classifiers.

    If the conversion of any of the classifiers is not possible, return None.
    Raise KeyError if the source data doesn't contain the found classifier.
    """

    if not TROVE2FEDORA_MAP:
        # Load the file with Python trove classifiers mapped to SPDX identifiers
        TROVE2FEDORA_MAP.update(_load_package_resource("classifiers_to_fedora.json"))

    spdx_identifiers = []
    for classifier in classifiers:
        try:
            fedora_identifier = TROVE2FEDORA_MAP[classifier]
        except KeyError as err:
            err_string = f"{classifier}: Such classifier doesn't exist. " \
            "If you believe that's pyp2spec's error, open an issue at the project's GitHub repo: " \
            "https://github.com/befeleme/pyp2spec"
            raise NoSuchClassifierError(err_string) from err

        # Classifiers that don't map unambiguously to a single Fedora SPDX expression are nulls
        # in the json
        # If the conversion is not possible, return None - meaning that another method of determining the license will be tried
        if fedora_identifier is None:
            return None
        spdx_identifiers.append(fedora_identifier)
    return spdx_identifiers


def license_keyword_to_spdx_identifiers(license_keyword: str | None) -> list | None:
    """Return a sorted list of SPDX identifiers extracted from the license_keyword
    
    If no identifiers were parsed out of the license_keyword or
    the license_keyword is unparseable, return None.
    """
    # nothing to transform, ergo no identifiers
    if not license_keyword:
        return None

    # `Artistic-1.0-Perl` alone is forbidden in Fedora, but the combination is allowed
    # These expressions may be a part of even longer license strings, which is impossible to cover
    # In an unlikely event of such license in the upstream data, raise an explicit exception
    special_cases = [
        "GPL-1.0-or-later OR Artistic-1.0-Perl",
        "GPL-2.0-or-later OR Artistic-1.0-Perl"
    ]
    for case in special_cases:
        if case in license_keyword:
            err_str = f"The detected license contains `{case}` which is handled in a special way in Fedora. " \
            "If you encountered this error, open an issue at https://github.com/befeleme/pyp2spec " \
            "and include the details about the package that causes the issue."
            raise NotImplementedError(err_str)

    licensing = get_spdx_licensing()
    try:
        parsed_license = licensing.parse(license_keyword, validate=True)
        # The objects are stored in sets, sort and return as a list
        return sorted(parsed_license.objects)
    except ExpressionError:
        # Don't bubble the error up, the calling function will handle the invalid result
        return None


def _load_fedora_licenses(
    source_path: str | None = None,
    session: Session | None = None
) -> dict[Any, Any]:
    """Load the dictionary of licenses evaluated for Fedora by the Fedora Legal team.

    Try to get them from the hard drive (installed by `fedora-license-data`).
    Fall back to the resources published by Fedora Legal online when the file isn't found.
    """

    fedora_licenses_path = source_path or "/usr/share/fedora-license-data/licenses/fedora-licenses.json"
    try:
        return _load_from_drive(fedora_licenses_path)

    except FileNotFoundError:
        url = "https://gitlab.com/fedora/legal/fedora-license-data/-/jobs/artifacts/main/raw/fedora-licenses.json?job=json"
        return _load_from_url(url, session=session)


def _is_compliant_with_fedora(identifier: str, fedora_licenses: dict[Any, Any]) -> bool:
    """Return True if the given identifier is "allowed" for Fedora and False if not.

    Fedora allows different types of licenses: "allowed for content", "allowed fonts",
    "allowed firmware", "allowed for documentation" and "allowed" for any use.
    For our purposes, only the generally "allowed" licenses are considered good.
    """

    # fedora_licenses doesn't have useful keys, hence iterate only through values
    # {"0": {
    #     "license": {
    #         "expression": "Rdisc",
    #         "status": ["allowed"],
    #         "url": "https://fedoraproject.org/wiki/Licensing/Rdisc_License"
    #     },
    #     "fedora": {
    #         "legacy-name": ["Rdisc License"],
    #         "legacy-abbreviation": ["Rdisc"]
    #     },
    #     "approved": "yes",
    #     "fedora_abbrev": "Rdisc",
    #     "fedora_name": "Rdisc License",
    #     "spdx_abbrev": "Rdisc"}...}

    for entry in fedora_licenses.values():
        license_info = entry.get("license")
        if license_info is not None:
            if identifier == entry["license"]["expression"]:
                return "allowed" in entry["license"]["status"]
    # No such identifier was found, we assume it's not good for Fedora
    return False


def check_compliance(
    license: str, *,
    licenses_dict: dict[Any, Any] | None = None,
    session: Session | None = None
) -> tuple[bool, dict[str, list[str]]]:

    """Determine whether the license is good for Fedora.

    Store the results in the checked_identifiers dictionary, under the
    respective `bad` and `good` keys.
    If no `spdx_identifiers` are given or not all are good, return tuple:
    (False, checked_identifies).
    Otherwise, return (True, checked_identifies).
    """

    # populate FEDORA_LICENSES only if they're still empty and
    # no other dictionary with licenses was given to be used here
    if licenses_dict is None and not FEDORA_LICENSES:
        FEDORA_LICENSES.update(_load_fedora_licenses(session=session))
    fedora_licenses = licenses_dict or FEDORA_LICENSES

    checked_identifies: dict[str, list[str]] = {
        "bad": [],
        "good": [],
    }

    spdx_identifiers = license_keyword_to_spdx_identifiers(license)
    if not spdx_identifiers:
        return (False, checked_identifies)
    for spdx_identifier in spdx_identifiers:
        if _is_compliant_with_fedora(spdx_identifier, fedora_licenses):
            checked_identifies["good"].append(spdx_identifier)
        else:
            checked_identifies["bad"].append(spdx_identifier)
    if checked_identifies["bad"]:
        return (False, checked_identifies)
    return (True, checked_identifies)


def transform_to_spdx(license_field: str | None, classifiers: list) -> tuple[list[str] | None, str | None]:
    """Return SPDX identifiers and expression based on the found
    package license metadata (classifiers or license keyword).

    If multiple identifiers are found, create an expression that's the safest option (with AND as joining operator).
    """

    if classifiers:
        identifiers = classifiers_to_spdx_identifiers(classifiers)
        if identifiers:
            expression = " AND ".join(identifiers)
            return (identifiers, expression)

    identifiers = license_keyword_to_spdx_identifiers(license_field)
    return (identifiers, license_field)


def generate_spdx_expression(license_field: str | None, classifiers: list) -> str | None:
    """Return the license expression based on detected metadata.

    If there are no identifiers, transformation to SPDX was unsuccessful.
    We don't want to process invalid license strings, so return None.
    Otherwise, we treat the expression as valid.
    """

    identifiers, expression = transform_to_spdx(license_field, classifiers)
    if not identifiers:
        return None
    return expression


def resolve_license_expression(data: RawMetadata | dict) -> str | None:
    if (expression := data.get("license_expression")):
        return expression
    return generate_spdx_expression(
        data.get("license"),
        filter_license_classifiers(data.get("classifiers", []))
    )



import pytest

from pyp2spec.license_processor import classifiers_to_spdx_identifiers, license_keyword_to_spdx_identifiers
from pyp2spec.license_processor import _is_compliant_with_fedora
from pyp2spec.license_processor import NoSuchClassifierError, check_compliance, generate_spdx_expression


@pytest.mark.parametrize(
    ("classifiers", "spdx_identifiers"),
    (
        (["License :: OSI Approved :: X.Net License"], ["Xnet"]),  # good
        (["License :: OSI Approved :: X.Net License", "License :: OSI Approved :: Artistic License"], None),  # good & bad
        (["License :: OSI Approved :: Artistic License"], None),  # bad
        (["License :: OSI Approved :: Eiffel Forum License", "License :: OSI Approved :: GNU General Public License (GPL)"], None),  # bad & bad
        (["License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)", "License :: OSI Approved :: ISC License (ISCL)"], ["LGPL-3.0-or-later", "ISC"]),  # good & good
    ),
)
def test_classifiers_to_spdx_identifiers(classifiers, spdx_identifiers):
    assert classifiers_to_spdx_identifiers(classifiers) == spdx_identifiers


@pytest.mark.parametrize(
    ("classifiers"),  (
        ["License :: OSI Approved :: XXXXXX"],
        ["Non-existing-classifier"],
        [""]
    ),
)
def test_conversion_to_spdx_fails_for_invalid_inputs(classifiers):
    with pytest.raises(NoSuchClassifierError):
        classifiers_to_spdx_identifiers(classifiers)


@pytest.mark.parametrize(
    ("expression", "expected"),
    (
        ("LGPL-2.1-only AND MIT AND BSD-2-Clause", ['BSD-2-Clause', 'LGPL-2.1-only', 'MIT']),
        ("MIT AND (LGPL-2.1-or-later OR BSD-3-Clause)", ['BSD-3-Clause', 'LGPL-2.1-or-later', 'MIT']),
        ("(MIT AND (LGPL-2.1+ AND BSD-3-Clause))", ['BSD-3-Clause', 'LGPL-2.1-or-later', 'MIT']),
        ("GPL-3.0-or-later WITH GCC-exception-3.1", ["GPL-3.0-or-later WITH GCC-exception-3.1"]),
        ("GPL-3.0-or-later WITH GCC-exception-3.1 OR Apache-2.0 OR MPL-2.0", ["Apache-2.0", "GPL-3.0-or-later WITH GCC-exception-3.1", "MPL-2.0"]),
        ("", None),
    ),
)
def test_split_expressions_to_identifiers(expression, expected):
    assert license_keyword_to_spdx_identifiers(expression) == expected


def test_perl_license_raises_exception():
    with pytest.raises(NotImplementedError):
        license_keyword_to_spdx_identifiers("MPL-2.0 OR GPL-2.0-or-later OR Artistic-1.0-Perl AND BSD-3-Clause")


@pytest.mark.parametrize(
    ("expression"),
    (
        ("FOO AND bar OR baz WITH EXCEPTION"),
        ("an arbitrary string"),
        ("multiline\nstring"),
        ("BSD BSD AND BSD"),
        ("string"),
    ),
)
def test_invalid_expressions_return_None(expression):
    assert license_keyword_to_spdx_identifiers(expression) is None



@pytest.mark.parametrize(
    ("identifier", "expected"),
    (
        ("CECILL-C", True),
        ("GPL-3.0-or-later WITH Classpath-exception-2.0", True),
        ("GPL-3.0-or-later WITH GCC-exception-3.1", True),
        ("CC0-1.0", False),  # allowed content license
        ("LicenseRef-LPPL", False),  # allowed font license
        ("EUPL-1.0", False),  # not allowed at all
        ("Bitstream-Vera", False),  # existing SPDX identifiers, but missing in fake_fedora_licenses
    )
)
def test_compliance_check_with_fedora(identifier, expected, fake_fedora_licenses):
    assert _is_compliant_with_fedora(identifier, fake_fedora_licenses) is expected


def test_no_license_classifiers_and_no_license_keyword():
    classifiers = []
    license_keyword = ""
    assert generate_spdx_expression(license_keyword, classifiers) is None


def test_license_from_multiple_classifiers():
    classifiers = [
        "License :: OSI Approved :: MIT License",
        "License :: OSI Approved :: MIT No Attribution License (MIT-0)",
    ]
    license_keyword = ""
    assert generate_spdx_expression(license_keyword, classifiers) == "MIT AND MIT-0"


def test_license_from_single_classifier():
    classifiers = ["License :: OSI Approved :: European Union Public Licence 1.0 (EUPL 1.0)"]
    license_keyword = ""
    assert generate_spdx_expression(license_keyword, classifiers) == "EUPL-1.0"


def test_mix_good_bad_classifiers():
    classifiers = [
        "License :: OSI Approved :: European Union Public Licence 1.0 (EUPL 1.0)",
        "License :: OSI Approved :: MIT License"
    ]
    license_keyword = ""
    assert generate_spdx_expression(license_keyword, classifiers) == "EUPL-1.0 AND MIT"


def test_license_keyword():
    license_keyword = "BSD-2-Clause"
    classifiers = []
    assert generate_spdx_expression(license_keyword, classifiers) == "BSD-2-Clause"


def test_license_keyword_and_classifiers():
    classifiers = [
        "License :: OSI Approved :: European Union Public Licence 1.0 (EUPL 1.0)",
        "License :: OSI Approved :: MIT License"
    ]
    license_keyword = "BSD-2-Clause"
    # classifiers always take precedence
    assert generate_spdx_expression(license_keyword, classifiers) == "EUPL-1.0 AND MIT"


@pytest.mark.parametrize(
    ("license", "good", "bad", "expected"),
    (
        ("MIT AND MIT-0", ["MIT", "MIT-0"], [], True),
        ("EUPL-1.0", [], ["EUPL-1.0"] , False),
        ("MIT AND EUPL-1.0", ["MIT"], ["EUPL-1.0"], False),
        ("RSCPL", [], ["RSCPL"], False),
        ("Fake-identifier", [], [], False),  # not a valid identifier, ergo not in "bad" list
        ("AAL AND MIT AND CECILL-B", ["AAL", "CECILL-B", "MIT"], [], True),
        ("", [], [], False),


    )
)
def test_check_compliance(license, good, bad, expected, fake_fedora_licenses):
    result, identifiers = check_compliance(license, licenses_dict=fake_fedora_licenses)
    assert result is expected
    assert identifiers["good"] == good
    assert identifiers["bad"] == bad



"""
This module takes care of loading all sorts of data from PyPI APIs.
"""
from __future__ import annotations
from typing import Any

from packaging.metadata import parse_email, RawMetadata
from packaging.version import Version
from requests import Response, Session

from pyp2spec.utils import Pyp2specError


class PackageNotFoundError(Pyp2specError):
    """Raised when there's no such package name on PyPI"""


class CoreMetadataNotFoundError(Pyp2specError):
    """Raised when there's no Metadata file available on PyPI API"""


class CompatibleVersionNotFoundError(Pyp2specError):
    """Raised when project doesn't have a version compatible with the requested one"""


def _get_from_url(url: str, error_str: str, session: Session | None = None) -> Response:
    _session = session or Session()
    response = _session.get(url)
    if not response.ok:
        raise PackageNotFoundError(error_str)
    return response


def _get_pypi_package_project_data(package: str, session: Session | None= None) -> dict[Any, Any]:
    pkg_index = f"https://pypi.org/pypi/{package}/json"
    error_str = f"Package `{package}` was not found on PyPI"
    return _get_from_url(pkg_index, error_str, session=session).json()


def _get_versioned_pypi_package_data(
    package: str,
    version: str, *,
    session: Session | None = None
) -> dict[Any, Any]:
    pkg_index = f"https://pypi.org/pypi/{package}/{version}/json"
    error_str = f"Package `{package}` or version `{version}` was not found on PyPI"
    return _get_from_url(pkg_index, error_str, session=session).json()


def _get_metadata_file(pypi_pkg_data: dict[Any, Any], session: Session | None = None) -> str:
    error_str = "The metadata file could not be located"
    for entry in pypi_pkg_data["urls"]:
        if entry["packagetype"] == "bdist_wheel":
            try:
                response = _get_from_url(entry["url"] + ".metadata", error_str, session=session)
            except PackageNotFoundError as exc:
                raise CoreMetadataNotFoundError(error_str) from exc
            return response.text
    else:
        raise CoreMetadataNotFoundError(error_str)


def _find_available_versions(project_releases: dict) -> list[str]:
    available_versions = []
    for release in project_releases:
        available_versions.append(release)
    return available_versions


def _find_compatible_version(compat: str, available_versions: list[str]) -> str | None:
    compatible_versions = [v for v in available_versions if v.startswith(compat)]
    if not compatible_versions:
        raise CompatibleVersionNotFoundError(f"There's no version compatible with the requested: `{compat}`")
    return max(compatible_versions, key=Version)


def load_from_pypi(
    package: str, *,
    version: str | None = None,
    compat: str | None = None,
    session: Session | None= None
) -> dict[Any, Any]:

    if version is None:
        pypi_project_data = _get_pypi_package_project_data(package, session=session)
        # Looking for the latest version
        if compat is None:
            version = pypi_project_data["info"]["version"]
        # Looking for the latest version of the compat version line
        else:
            available_versions = _find_available_versions(pypi_project_data["releases"])
            version = _find_compatible_version(compat, available_versions)

    return _get_versioned_pypi_package_data(package, version=version, session=session)


def load_core_metadata_from_pypi(pypi_pkg_data: dict[Any, Any], session: Session | None = None) -> RawMetadata:
    metadata = _get_metadata_file(pypi_pkg_data, session=session)
    raw, _ = parse_email(metadata)
    # TODO: consider porting to packaging.Metadata instance?
    return raw



"""Test the functionality of pyp2conf part of the tool.
The data is downloaded from the PyPI and stored in betamax cassettes
to prevent loading from the internet on each request.
"""
import pytest

try:
    import tomllib
except ImportError:
    import tomli as tomllib

from pyp2spec.pyp2conf import create_config_contents, prepare_package_info, check_compliance, gather_package_info
from pyp2spec.pypi_loaders import PackageNotFoundError


def test_non_existent_package(betamax_session):
    with pytest.raises(PackageNotFoundError):
        create_config_contents(
            {"package": "definitely-nonexisting-package-name"},
            session=betamax_session)


@pytest.mark.parametrize("package, version",
    [
        ("aionotion", "2.0.3"),
        ("click", "8.1.7"),
    ]
)
def test_automatically_generated_config_is_valid(betamax_parametrized_session, package, version):
    """Run the config rendering in fully automated mode and compare the results"""
    config = create_config_contents(
        {"package": package,
        "version": version},
        session=betamax_parametrized_session,
    )

    with open(f"tests/test_configs/default_python-{package}.conf", "rb") as config_file:
        loaded_contents = tomllib.load(config_file)

    assert config == loaded_contents


@pytest.mark.parametrize("package, version, alt_python",
    [
        ("Pello", "1.0.4", "3.9"),
    ]
)
def test_automatically_generated_config_with_alt_python_is_valid(
        betamax_parametrized_session, package, version, alt_python
    ):
    """Run the config rendering in fully automated mode and compare the results"""
    config = create_config_contents(
        {"package": package,
        "version": version,
        "python_alt_version": alt_python},
        session=betamax_parametrized_session,
    )

    with open(f"tests/test_configs/default_python{alt_python}-{package.lower()}.conf", "rb") as config_file:
        loaded_contents = tomllib.load(config_file)

    assert config == loaded_contents


@pytest.mark.parametrize("package, compat, version",
    [
        ("pytest", "7", None),
        ("pytest", "7.2", "7.2.1")  # not the first, nor latest
    ]
)
def test_automatically_generated_compat_config_is_valid(
        betamax_parametrized_session, package, compat, version
    ):
    config = create_config_contents(
        {"package": package,
        "version": version,
        "compat": compat},
        session=betamax_parametrized_session,
    )

    with open(f"tests/test_configs/default_python-{package}{compat}.conf", "rb") as config_file:
        loaded_contents = tomllib.load(config_file)
    assert config["compat"] == compat
    assert config == loaded_contents


def test_config_with_customization_is_valid(betamax_session):
    """Get the upstream metadata and modify some fields to get the custom config file.
    This also tests the compliance with Fedora Legal data by making
    a request to the remote resource.
    """
    config = create_config_contents(
        {"package": "aionotion",
        "version": "2.0.3",
        "automode": True,
        "compliant": True},
        session=betamax_session,
    )

    with open(f"tests/test_configs/customized_aionotion.conf", "rb") as config_file:
        loaded_contents = tomllib.load(config_file)

    assert config == loaded_contents


def test_archful_package(betamax_session):
    """Generate config for numpy which is archful"""
    config = create_config_contents(
        {"package": "numpy",
        "version": "1.25.2",
        "automode": True},
        session=betamax_session,
    )

    with open(f"tests/test_configs/default_python-numpy.conf", "rb") as config_file:
        loaded_contents = tomllib.load(config_file)

    assert config["archful"] == loaded_contents["archful"]
    assert config == loaded_contents


def test_package_with_extras(betamax_session):
    config = create_config_contents(
        {"package": "sphinx"},
        session=betamax_session,
    )

    with open(f"tests/test_configs/customized_python-sphinx.conf", "rb") as config_file:
        loaded_contents = tomllib.load(config_file)

    assert config["extras"] == loaded_contents["extras"]
    assert config == loaded_contents


def test_no_license_classifiers_and_no_license_keyword():
    fake_pkg_data = {
        "name": "foo",
        "classifiers": [],
        "license_keyword": "",
    }
    pkg = prepare_package_info(fake_pkg_data)
    assert pkg.license is None


def test_license_expression_takes_precedence(fake_fedora_licenses):
    fake_pkg_data = {
        "name": "foo",
        "classifiers": ["License :: OSI Approved :: MIT License"],
        "license_keyword": "MIT-0",
        "license_expression": "BSD-2-Clauseo",
    }
    pkg = prepare_package_info(fake_pkg_data)
    assert pkg.license == "BSD-2-Clauseo"
    result, identifiers = check_compliance(pkg.license, licenses_dict=fake_fedora_licenses)
    assert not result
    assert not identifiers["good"]
    # There are none valid identifiers
    assert not identifiers["bad"]


def test_mix_good_bad_licenses_fail_compliance_check(fake_fedora_licenses):
    fake_pkg_data = {
        "name": "foo",
        "classifiers": [
            "License :: OSI Approved :: European Union Public Licence 1.0 (EUPL 1.0)",
            "License :: OSI Approved :: MIT License"
        ],
    }
    pkg = prepare_package_info(fake_pkg_data)
    assert pkg.license == "EUPL-1.0 AND MIT"
    result, identifiers = check_compliance(pkg.license, licenses_dict=fake_fedora_licenses)
    assert not result
    assert identifiers["good"] == ["MIT"]
    assert identifiers["bad"] == ["EUPL-1.0"]


def test_summary_is_generated_if_not_in_upstream():
    fake_pkg_data = {
        "name": "foo",
        "summary": "",
    }
    pkg = prepare_package_info(fake_pkg_data)
    assert pkg.summary == "..."


def test_summary_is_generated_if_upstream_data_is_multiline():
    fake_pkg_data = {
        "name": "foo",
        "summary": "I\nforgot\nthat summary\nmust\nbe short",
    }
    pkg = prepare_package_info(fake_pkg_data)
    assert pkg.summary == "..."


def test_capitalized_underscored_pypi_name_is_normalized():
    fake_pkg_data = {
        "name": "Awesome_TestPkg",
    }
    pkg = prepare_package_info(fake_pkg_data)
    assert pkg.pypi_name == "awesome-testpkg"


@pytest.mark.parametrize(
    ("metadata", "lf_present"), [
        (["License.rst"], True),
        (["License.rst", "License.md"], True),
        ([], False),
    ]
)
def test_license_files_in_metadata_files(metadata, lf_present):
    fake_pkg_data = {
        "name": "foo",
        "license_files": metadata,
    }
    pkg = prepare_package_info(fake_pkg_data)
    assert pkg.license_files_present is lf_present


def test_prepare_package_info_pypi_source():
    data = {"info": {
        "name": "example",
        "summary": "A sample project",
        "license_expression": "MIT",
        "license": "MIT License",
        "classifiers": ["License :: OSI Approved :: MIT License"],
        "requires_dist": ["requests"],
        "version": "1.0.0",
        "license_files": ["LICENSE"],
        "project_urls": {"Homepage": "https://example.com"},
        "yanked": False,
        "maintainer": "John Doe",
    }, "releases": [],}
    result = prepare_package_info(data["info"])
    assert result.pypi_name == "example"
    assert result.summary == "A sample project"
    assert result.license_files_present is True
    assert result.license == "MIT"
    assert result.extras == []
    assert result.pypi_version == "1.0.0"
    assert result.url == "https://example.com"


def test_prepare_package_info_core_metadata():
    data = {
        "name": "example",
        "summary": "A sample project",
        "license_expression": "MIT",
        "license": "MIT License",
        "classifiers": ["License :: OSI Approved :: MIT License"],
        "requires_dist": ["requests"],
        "version": "1.0.0",
        "license_files": ["LICENSE"],
        "home_page": "https://example.com",
        "metadata_version": "2.1",
    }
    result = prepare_package_info(data)
    assert result.pypi_name == "example"
    assert result.summary == "A sample project"
    assert result.license_files_present is True
    assert result.license == "MIT"
    assert result.extras == []
    assert result.pypi_version == "1.0.0"
    assert result.url == "https://example.com"


def test_prepare_package_info_missing_keys():
    data = {
        "name": "foo",
    }
    result = prepare_package_info(data)
    assert result.pypi_name == "foo"
    assert result.summary == "..."
    assert result.license_files_present is False
    assert result.license is None
    assert result.extras == []
    assert result.pypi_version == ""
    assert result.url == "..."


def test_gather_package_info_pypi_source():
    pypi = {"info": {
        "name": "example",
        "summary": "A sample project",
        "license_expression": "MIT",
        "license": "MIT License",
        "classifiers": ["License :: OSI Approved :: MIT License"],
        "requires_dist": ["requests"],
        "version": "1.0.0",
        "license_files": ["LICENSE"],
        "project_urls": {"Homepage": "https://example.com"},
        "yanked": False,
        "maintainer": "John Doe",
        }, "releases": [],
        "urls": [{
            "packagetype": "sdist",
            "filename": "example-0.2-tar.gz"
        }, {
            "packagetype": "bdist_wheel",
            "filename": "example-0.2-py2.py3-none-any.whl"
        }]}
    result = gather_package_info(None, pypi)
    assert result.archive_name == "example-0.2-tar.gz"
    assert result.archful is False


def test_gather_package_info_core_metadata():
    data = {
        "name": "example",
        "summary": "A sample project",
        "license_expression": "MIT",
        "license": "MIT License",
        "classifiers": ["License :: OSI Approved :: MIT License"],
        "requires_dist": ["requests"],
        "version": "1.0.0",
        "license_files": ["LICENSE"],
        "home_page": "https://example.com",
        "metadata_version": "2.1",
    }
    pypi = {"info": {
        "name": "discarded",
        "summary": "Deliberately different metadata",
        "license_expression": "BSD",
        "license": "BSD",
        "classifiers": ["License :: OSI Approved :: BSD License"],
        "requires_dist": ["click"],
        "version": "7.0.0",
        "license_files": ["LICENSE.txt"],
        "project_urls": {"Source": "https://example.com"},
        "yanked": False,
        "maintainer": "John Doe",
        }, "releases": [],
        "urls": [{
            "packagetype": "sdist",
            "filename": "example-7.0-tar.gz"
        }, {"packagetype": "bdist_wheel",
            "filename": "example-7.0-cp34-abi3-manylinux1_x86_64.whl"
        }]}
    result = gather_package_info(data, pypi)
    assert result.pypi_name == "example"
    assert result.summary == "A sample project"
    assert result.license_files_present is True
    assert result.license == "MIT"
    assert result.extras == []
    assert result.pypi_version == "1.0.0"
    assert result.url == "https://example.com"
    assert result.archive_name == "example-7.0-tar.gz"
    assert result.archful is True



"""Test the functionality of functions getting the data from PyPI.
The data is downloaded and stored in betamax cassettes
to prevent loading from the internet on each request.
"""

import pytest

from pyp2spec.pypi_loaders import load_from_pypi, load_core_metadata_from_pypi
from pyp2spec.pypi_loaders import PackageNotFoundError, CompatibleVersionNotFoundError
from pyp2spec.pypi_loaders import _find_available_versions, _find_compatible_version


def test_load_from_pypi_no_version_given(betamax_session):
    result = load_from_pypi("Sphinx", session=betamax_session)
    assert result["info"]["version"] == "8.1.3"


def test_load_from_pypi_particular_version(betamax_session):
    result = load_from_pypi("Sphinx", version="7.3.7", session=betamax_session)
    assert result["info"]["version"] == "7.3.7"


def test_load_from_pypi_package_not_found(betamax_session):
    package = "non-existent-package"
    with pytest.raises(PackageNotFoundError):
        load_from_pypi(package, session=betamax_session)


def test_load_core_metadata(betamax_session):
    pypi_pkg_data = load_from_pypi("Sphinx", session=betamax_session)
    result = load_core_metadata_from_pypi(pypi_pkg_data, betamax_session)
    assert result["name"] == "Sphinx"
    assert result["version"] == "8.1.3"
    assert result["provides_extra"] == ["docs", "lint", "test"]



{"2.0.0":["..."],"2.0.1":["..."],"3.0.0":["..."],"6.0.1":["..."],"2.1.0":["..."],"2.1.1":["..."]}
def test_find_available_versions():
    releases = {
        "2.0.0":["..."],"2.0.1":["..."],"3.0.0":["..."],
        "6.0.1":["..."],"2.1.0":["..."],"2.1.1":["..."]
    }
    expected = ["2.0.0", "2.0.1", "3.0.0", "6.0.1", "2.1.0", "2.1.1"]
    assert _find_available_versions(releases) == expected


@pytest.mark.parametrize(
    ("compat", "available_versions", "expected"), [
        ("7", ["6.0.0", "7.0.0", "7.9.1"], "7.9.1"),
        ("7.4", ["7.4.2", "6.0.0", "7.4.3", "7.0.0", "7.0.1"], "7.4.3"),
        ("7.2", ["7.4.2", "7.2.3", "7.4.3", "6.0.0", "7.0.1"], "7.2.3"),
        ("7.4.2", ["7.4.2", "7.2.3", "7.4.3", "6.0.0", "7.0.1"], "7.4.2"),
    ]
)
def test_find_compatible_version(compat, available_versions, expected):
    assert _find_compatible_version(compat, available_versions) == expected


def test_no_compatible_version_available():
    with pytest.raises(CompatibleVersionNotFoundError):
        _find_compatible_version("9", ["6.0.0", "7.0.0", "7.9.1"])
